# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-13 11:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0069_similarproductspluginmodel'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='slug',
            field=models.SlugField(unique=True, verbose_name='Slug'),
        ),
        migrations.AlterField(
            model_name='product',
            name='slug',
            field=models.SlugField(max_length=255, unique=True, verbose_name='Slug'),
        ),
        migrations.AlterField(
            model_name='propertyvalueslug',
            name='slug',
            field=models.SlugField(max_length=255, unique=True, verbose_name='Slug'),
        ),
        migrations.AlterField(
            model_name='vendor',
            name='slug',
            field=models.SlugField(max_length=255, unique=True, verbose_name='Slug'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-30 11:45
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0065_auto_20170930_1249'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionpluginmodel',
            name='limit',
            field=models.PositiveIntegerField(default=6, verbose_name='Limit'),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from django.conf import settings


_PCART_COLLECTION_ORDERINGS = {
    'cheap': {
        'label': _('From cheaper to more expensive'),
        'without_filters': ['-status__weight', 'min_variant_price', 'id'],
        'with_filters': ['-status__weight', 'filtered_min_variant_price', 'id'],
    },
    'expensive': {
        'label': _('From more expensive to cheaper'),
        'without_filters': ['-status__weight', '-max_variant_price', 'id'],
        'with_filters': ['-status__weight', '-filtered_max_variant_price', 'id'],
    },
}

PCART_COLLECTION_ORDERINGS = getattr(settings, 'PCART_COLLECTION_ORDERINGS', _PCART_COLLECTION_ORDERINGS)

PCART_COLLECTION_DEFAULT_ORDERING = getattr(settings, 'PCART_COLLECTION_DEFAULT_ORDERING', 'cheap')

PCART_ENABLE_PRODUCT_VARIANTS = getattr(settings, 'PCART_ENABLE_PRODUCT_VARIANTS', True)



from django.db import models
from django.contrib.postgres.fields import JSONField, ArrayField
from django.utils.translation import ugettext_lazy as _
from django.contrib.sites.models import Site
from django.urls import reverse, NoReverseMatch
from django.db.models.signals import pre_save, post_save, post_delete
from django.dispatch import receiver
from django.conf import settings
# from django.utils.functional import lazy
from django.core.cache import cache
from adminsortable.fields import SortableForeignKey
from adminsortable.models import SortableMixin
from mptt.models import MPTTModel, TreeForeignKey
from mptt.managers import TreeManager
from mptt.querysets import TreeQuerySet
from cms.models import CMSPlugin
from cms.models.fields import PlaceholderField
from djangocms_text_ckeditor.fields import HTMLField
from parler.models import TranslatableModel, TranslatedFields
from parler.managers import TranslatableManager, TranslatableQuerySet
import uuid

from typing import List, Tuple, Union, Optional, Dict


MAX_TAG_LENGTH = 30


class ProductType(TranslatableModel):
    """
    Describes a nature of any product from the catalog. Is using as a container for a properties list.
    The `title` value must be unique across the database.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255, unique=True),
        # SEO (you can store templates here)
        # Usable for setup SEO templates for all products of the particular type.
        page_title=models.CharField(_('Page title'), max_length=255, blank=True),
        meta_description=models.TextField(_('Meta description'), blank=True),
        meta_keywords=models.TextField(_('Meta keywords'), blank=True),
        meta_text=models.TextField(_('Meta text'), blank=True),
    )

    # Useful default values
    default_quantity_for_purchase = models.PositiveIntegerField(
        _('Default quantity for purchase'), default=1)

    # CMS placeholders
    primary_placeholder = PlaceholderField(
        'primary_product_type_placeholder', related_name='product_type_primary')
    secondary_placeholder = PlaceholderField(
        'secondary_product_type_placeholder', related_name='product_type_secondary')

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product type')
        verbose_name_plural = _('Product types')
        # ordering = ['title']

    def __str__(self) -> str:
        return self.title

    def get_properties_fields(self, initial_values: Optional[Dict] = None, for_variants: bool = False) -> List[Tuple]:
        """
        Returns the list of tuples with some data useful for the product managing forms for admin interface.
        """
        from django import forms

        if initial_values is None:
            initial_values = dict()  # A trick for prevent using the mutable value as default

        if for_variants:
            properties = self.properties.filter(for_variants=True)
        else:
            properties = self.properties.all()
        result = [
            (
                'property_%s' % i,
                p.title,
                forms.CharField(label=p.title, initial=initial_values.get(p.title) or p.default_value, required=False),
            ) for i, p in enumerate(properties)]
        return result

    def get_visible_properties(self):
        _result = self.properties.exclude(hidden=True)
        return _result


class ProductTypeProperty(SortableMixin):
    """
    Represents a product property related to the specified product type.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(_('Title'), max_length=255)

    default_value = models.CharField(_('Default value'), max_length=255, blank=True)
    tag_prefix = models.CharField(_('Tag prefix'), max_length=10, blank=True)

    use_in_filters = models.BooleanField(_('Use in filters'), default=False)
    for_variants = models.BooleanField(_('For variants'), default=False)

    hidden = models.BooleanField(
        _('Hidden'), default=False,
        help_text=_('Useful for hide some product specified information from a web page.'))

    product_type = SortableForeignKey(ProductType, verbose_name=_('Product type'), related_name='properties')

    position = models.PositiveIntegerField(_('Position'), default=0, editable=False, db_index=True)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product type property')
        verbose_name_plural = _('Product type properties')
        ordering = ['position']

    def __str__(self) -> str:
        return self.title


# Listen for signals for ProductTypeProperty

@receiver(pre_save, sender=ProductTypeProperty)
def product_type_pre_save_listener(sender, instance: 'ProductTypeProperty', **kwargs):
    _updated = False
    if instance.pk:
        from .tasks import rename_property_name, update_product_tags
        product_type_id = instance.product_type_id
        try:
            ptp = ProductTypeProperty.objects.get(pk=instance.pk)
            old_title = ptp.title
            new_title = instance.title
            if old_title != new_title:
                rename_property_name.delay(product_type_id, old_title, new_title)
                _updated = True
            if not _updated and ptp.tag_prefix != instance.tag_prefix:
                update_product_tags.delay(product_type_id, new_title)
        except ProductTypeProperty.DoesNotExist:
            pass


def generate_collection_image_filename(instance: 'Collection', filename: str) -> str:
    ext = filename.split('.')[-1]
    url = 'images/collections/%s/%s.%s' % (instance.id, str(uuid.uuid4()).replace('-', ''), ext)
    return url


class CollectionQuerySet(TranslatableQuerySet, TreeQuerySet):
    pass
    # Optional: make sure the Django 1.7 way of creating managers works.
    # def as_manager(cls):
    #     manager = CategoryManager.from_queryset(cls)()
    #     manager._built_with_as_manager = True
    #     return manager
    # as_manager.queryset_only = True
    # as_manager = classmethod(as_manager)


class CollectionManager(TreeManager, TranslatableManager):
    queryset_class = CollectionQuerySet

    def get_queryset(self):
        # This is the safest way to combine both get_queryset() calls
        # supporting all Django versions and MPTT 0.7.x versions
        return self.queryset_class(self.model, using=self._db).order_by(self.tree_id_attr, self.left_attr)


class Collection(MPTTModel, TranslatableModel):
    """
    Describes a page on the site with a list of products and such features like filtering, sorting and
    support of multiple templates.
    """

    DEFAULT_URL_FILTER_RULES = '''VENDOR
PRICE'''
    PRODUCTS_SELECTION_CHOICES = (
        ('manual', _('Manual')),
        ('automatic', _('Automatic')),
    )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    site = models.ForeignKey(
        Site, verbose_name=_('Site'), related_name='collections', on_delete=models.PROTECT)
    slug = models.SlugField(_('Slug'), unique=True)
    parent = TreeForeignKey(
        'self', null=True, blank=True, related_name='children', db_index=True,
        verbose_name=_('Parent'),
    )

    image = models.ImageField(_('Image'), null=True, blank=True, upload_to=generate_collection_image_filename)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255),
        description=HTMLField(_('Description'), blank=True),
        # SEO
        page_title=models.CharField(_('Page title'), max_length=255, blank=True),
        meta_description=models.TextField(_('Meta description'), blank=True),
        meta_keywords=models.TextField(_('Meta keywords'), blank=True),
        meta_text=models.TextField(_('Meta text'), blank=True),
    )

    # URL filter rules
    custom_url_filter_rules = models.BooleanField(_('Custom URL filter rules'), default=False)
    url_filter_rules = models.TextField(
        _('URL filter rules'), blank=True, default=DEFAULT_URL_FILTER_RULES,
        help_text=_('Use separate lines for different rules. See documentation for details.')
    )

    # Filters visibility
    show_vendor_filter = models.BooleanField(_('Show vendor filter'), default=True)
    show_properties_filters = models.BooleanField(_('Show properties filters'), default=True)
    exclude_properties_filters = ArrayField(
        models.CharField(max_length=70),
        verbose_name=_('Exclude properties filters'), default=list, blank=True,
        help_text=_('Comma separated list of properties labels which you want to exclude from filters.'),
    )
    show_price_filter = models.BooleanField(_('Show price filter'), default=True)

    # Product selection mode
    products_selection = models.CharField(
        _('Products selection'), max_length=30, default='manual',
        choices=PRODUCTS_SELECTION_CHOICES,
        help_text=_('Use Automatic mode if you want to select products based on conditions.'),
    )

    # The list of conditions for automated collections
    condition_vendors = models.ManyToManyField(
        'Vendor', verbose_name=_('Vendors'), blank=True, related_name='auto_collections')
    condition_product_types = models.ManyToManyField(
        'ProductType', verbose_name=_('Product types'), blank=True, related_name='auto_collections')
    condition_min_price = models.DecimalField(_('Min price'), max_digits=10, decimal_places=2, default=0.00)
    condition_max_price = models.DecimalField(_('Max price'), max_digits=10, decimal_places=2, default=0.00)
    mandatory_tags = ArrayField(
        models.CharField(max_length=MAX_TAG_LENGTH), verbose_name=_('Mandatory tags'),
        default=list, blank=True,
    )
    optional_tags = ArrayField(
        models.CharField(max_length=MAX_TAG_LENGTH), verbose_name=_('Optional tags'),
        default=list, blank=True,
    )

    # Publication
    published = models.BooleanField(_('Published'), default=True)

    # CMS placeholders
    primary_placeholder = PlaceholderField(
        'primary_collection_placeholder', related_name='collection_primary')
    secondary_placeholder = PlaceholderField(
        'secondary_collection_placeholder', related_name='collection_secondary')
    extra_placeholder = PlaceholderField(
        'secondary_collection_placeholder', related_name='collection_extra')

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    objects = CollectionManager()

    class Meta:
        verbose_name = _('Collection')
        verbose_name_plural = _('Collections')

    def __str__(self) -> str:
        return self.title

    def get_absolute_url(self) -> str:
        try:
            return reverse('pcart_collection:product-list-for-collection', args=[self.slug])
        except NoReverseMatch:
            return '#no-page-for-collection-app'

    def as_dict(self) -> Dict:
        result = {
            'id': self.id,
            'title': self.title,
            'site': {'domain': self.site.domain, 'name': self.site.name},
            'slug': self.slug,
            'parent': self.parent_id,
            'description': self.description,
            'page_title': self.page_title,
            'published': self.published,
            'added': self.added,
            'changed': self.changed,
            'url': self.get_absolute_url(),
        }
        return result

    def get_product_types(self):
        """ Returns a set of unique ProductType instances available for the collection."""
        if self.products_selection == 'manual':
            return ProductType.objects.filter(products__collections=self).distinct()
        elif self.products_selection == 'automatic':
            return self.condition_product_types.all()

    def get_url_filter_rules(self) -> str:
        """ Returns URL filter rules"""
        if self.custom_url_filter_rules:
            return self.url_filter_rules
        else:
            _cache_key = 'url-filter-rules-%s' % self.pk
            result = cache.get(_cache_key)
            if result:
                return result

            _product_types = self.get_product_types()
            _filter_rules = []
            if self.show_vendor_filter:
                _filter_rules.append('VENDOR')
            if self.show_properties_filters:
                for p_type in _product_types:
                    f_props = p_type.properties.filter(use_in_filters=True)
                    for prop in f_props:
                        if prop.title not in self.exclude_properties_filters:
                            _filter_rules.append('TAG/+%s/VALUE ONLY IGNORE DASH' % prop.tag_prefix)
            if self.show_price_filter:
                _filter_rules.append('PRICE')
            result = '\n'.join(_filter_rules)
            cache.set(_cache_key, result, 60*15)    # save for 15 min
            return result


class ProductStatus(TranslatableModel):
    """
    Represents a product status model.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    # title = models.CharField(_('Title'), max_length=255, unique=True)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255, unique=True)
    )

    show_buy_button = models.BooleanField(_('Show buy button'), default=True)
    is_visible = models.BooleanField(_('Is visible'), default=True)
    is_searchable = models.BooleanField(_('Is searchable'), default=True)

    weight = models.PositiveIntegerField(_('Weight'), default=0)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product status')
        verbose_name_plural = _('Product statuses')
        ordering = ['-weight']

    def __str__(self) -> str:
        return self.title

    def as_dict(self) -> Dict:
        result = {
            'id': self.id,
            'title': self.title,
            'show_buy_button': self.show_buy_button,
            'is_visible': self.is_visible,
            'is_searchable': self.is_searchable,
            'weight': self.weight,
        }
        return result


def generate_vendor_image_filename(instance: 'Vendor', filename: str) -> str:
    ext = filename.split('.')[-1]
    url = 'images/vendors/%s/%s.%s' % (instance.id, str(uuid.uuid4()).replace('-', ''), ext)
    return url


class Vendor(TranslatableModel):
    """
    Represents a product vendor.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(_('Slug'), max_length=255, unique=True)
    image = models.ImageField(_('Image'), null=True, blank=True, upload_to=generate_vendor_image_filename)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255, unique=True),
        description=HTMLField(_('Description'), blank=True),
    )

    class Meta:
        verbose_name = _('Vendor')
        verbose_name_plural = _('Vendors')

    def __str__(self) -> str:
        return self.title

    def as_dict(self) -> dict:
        result = {
            'id': self.id,
            'title': self.title,
        }
        return result


class Product(TranslatableModel):
    """
    Represents a product.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(_('Slug'), max_length=255, unique=True)
    external_id = models.CharField(
        _('External ID'), default='', blank=True, max_length=255, db_index=True)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255),
        description=HTMLField(_('Description'), blank=True),
        # SEO
        page_title=models.CharField(_('Page title'), max_length=255, blank=True),
        meta_description=models.TextField(_('Meta description'), blank=True),
        meta_keywords=models.TextField(_('Meta keywords'), blank=True),
        meta_text=models.TextField(_('Meta text'), blank=True),
    )

    vendor = models.ForeignKey(
        Vendor, verbose_name=_('Vendor'), related_name='products', null=True, blank=True, on_delete=models.SET_NULL)
    product_type = models.ForeignKey(
        ProductType, verbose_name=_('Product type'), related_name='products', on_delete=models.CASCADE)
    collections = models.ManyToManyField(
        Collection, verbose_name=_('Collections'), related_name='products', blank=True,
        help_text=_('Add this product to a collection so it\'s easy to find in your store.'),
    )

    tags = ArrayField(
        models.CharField(max_length=MAX_TAG_LENGTH),
        verbose_name=_('Tags'),
        blank=True,
        default=list,
    )
    properties = JSONField(_('Properties'), default=dict, blank=True)

    sku = models.CharField(_('SKU (Stock Keeping Unit)'), blank=True, max_length=100)
    barcode = models.CharField(_('Barcode (ISBN, UPC, GTIN, etc.)'), blank=True, max_length=100)

    status = models.ForeignKey(
        ProductStatus, verbose_name=_('Status'), related_name='products', on_delete=models.PROTECT)
    quantity = models.PositiveIntegerField(_('Quantity'), default=None, null=True, blank=True)

    price = models.DecimalField(_('Price'), max_digits=10, decimal_places=2, default=0.00)
    compare_at_price = models.DecimalField(_('Compare at price'), max_digits=10, decimal_places=2, default=0.00)
    price_varies = models.BooleanField(_('Price varies'), default=False)

    # Set automatically with stored procedure. For read only access.
    max_variant_price = models.DecimalField(_('Max variant price'), max_digits=10, decimal_places=2, default=0.00)
    min_variant_price = models.DecimalField(_('Min variant price'), max_digits=10, decimal_places=2, default=0.00)
    variants_count = models.PositiveIntegerField(_('Variants count'), default=0)

    weight = models.FloatField(
        _('Weight (kg)'), default=0.0, help_text=_('Used to calculate shipping rates at checkout.'))

    # Some extra attributes
    rank_1 = models.FloatField(_('Rank 1'), default=0.0)
    rank_2 = models.FloatField(_('Rank 2'), default=0.0)
    rank_3 = models.FloatField(_('Rank 3'), default=0.0)

    extra_int_1 = models.IntegerField(_('Extra int 1'), default=0)
    extra_int_2 = models.IntegerField(_('Extra int 2'), default=0)
    extra_float_1 = models.FloatField(_('Extra float 1'), default=0.0)
    extra_float_2 = models.FloatField(_('Extra float 2'), default=0.0)

    extra_price_1 = models.DecimalField(_('Extra price 1'), max_digits=10, decimal_places=2, default=0.00)
    extra_price_2 = models.DecimalField(_('Extra price 2'), max_digits=10, decimal_places=2, default=0.00)
    extra_price_3 = models.DecimalField(_('Extra price 3'), max_digits=10, decimal_places=2, default=0.00)

    extra_text_1 = models.TextField(_('Extra text 1'), default='', blank=True)
    extra_text_2 = models.TextField(_('Extra text 2'), default='', blank=True)

    extra_data = JSONField(_('Extra data'), default=dict, blank=True)
    extra_id_1 = models.CharField(_('Extra ID 1'), default='', blank=True, max_length=255, db_index=True)
    extra_id_2 = models.CharField(_('Extra ID 2'), default='', blank=True, max_length=255, db_index=True)

    published = models.BooleanField(_('Published'), default=True)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product')
        verbose_name_plural = _('Products')
        ordering = ['id']

    def __str__(self) -> str:
        return self.title

    @staticmethod
    def type() -> str:
        return 'product'

    def get_absolute_url(self) -> str:
        try:
            return reverse('pcart_product:product-detail', args=[self.slug])
        except NoReverseMatch:
            return '#no-page-for-product-app'

    def get_properties_fields(self, initial_values: Optional[Dict] = None, for_variants: bool = False) -> List[Tuple]:
        """
        Returns properties list for generate product form for admin.
        """
        return self.product_type.get_properties_fields(
            initial_values=self.properties if initial_values is None else initial_values,
            for_variants=for_variants)

    def has_variants(self) -> bool:
        """
        Returns `True` if the product has one or more variants.
        """
        return self.variants_count > 0

    def get_image(self) -> 'ProductImage':
        """
        Returns a first available image for the product.
        """
        return self.images.first()

    def available(self) -> bool:
        """
        Returns `True` if the product does not have variants, positive price and corresponding status.
        """
        return self.variants_count == 0 and self.price > 0 and self.status.show_buy_button

    def as_dict(self) -> Dict:
        result = {
            'id': self.id,
            'title': self.title,
            'slug': self.slug,
            'description': self.description,
            'page_title': self.page_title,
            'tags': self.tags,
            'properties': self.properties,
            'sku': self.sku,
            'barcode': self.barcode,
            'status': self.status.as_dict(),
            'price': self.price,
            'price_varies': self.price_varies,
            'compare_at_price': self.compare_at_price,
            'max_variant_price': self.max_variant_price,
            'min_variant_price': self.min_variant_price,
            'variants_count': self.variants_count,
            'weight': self.weight,
            'published': self.published,
            'url': self.get_absolute_url(),
        }
        return result

    def get_page_title(self) -> str:
        """
        Returns a product page title.
        """
        return self.page_title or self.title

    def get_collections(self):
        from django.db.models import Q
        manual_collections_ids = self.collections.values_list('id', flat=True)
        _result = Collection.objects.filter(
            Q(pk__in=manual_collections_ids) | (
                Q(products_selection='automatic') &
                (Q(condition_product_types=None) | Q(condition_product_types=self.product_type)) &
                (Q(condition_vendors=None) | Q(condition_vendors=self.vendor)) &
                (Q(condition_min_price=0.0) | Q(condition_min_price__lte=self.price)) &
                (Q(condition_max_price=0.0) | Q(condition_max_price__gte=self.price)) &
                (Q(mandatory_tags=[]) | Q(mandatory_tags__contained_by=self.tags)) &
                (Q(optional_tags=[]) | Q(optional_tags__overlap=self.tags))
            )
        ).prefetch_related('translations')
        return _result


@receiver(post_save, sender=Product)
def product_post_save_listener(sender, instance: 'Product', **kwargs):
    from .utils import filter_non_numeric_tags
    tagset = instance.tags
    values = filter_non_numeric_tags(tagset)
    PropertyValueSlug.objects.add_values(values)


class ProductVariant(TranslatableModel):
    """
    Represents a product variant.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(_('Slug'), unique=False, max_length=255)
    external_id = models.CharField(
        _('External ID'), default='', blank=True, max_length=255, db_index=True)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255)
    )

    product = models.ForeignKey(
        Product, verbose_name=_('Product'), related_name='variants', on_delete=models.CASCADE)

    tags = ArrayField(
        models.CharField(max_length=MAX_TAG_LENGTH),
        verbose_name=_('Tags'),
        blank=True,
        default=list,
    )
    properties = JSONField(_('Properties'), default=dict, blank=True)

    sku = models.CharField(_('SKU (Stock Keeping Unit)'), blank=True, max_length=100)
    barcode = models.CharField(_('Barcode (ISBN, UPC, GTIN, etc.)'), blank=True, max_length=100)

    status = models.ForeignKey(ProductStatus, verbose_name=_('Status'), related_name='variants')
    quantity = models.PositiveIntegerField(_('Quantity'), default=None, null=True, blank=True)

    price = models.DecimalField(_('Price'), max_digits=10, decimal_places=2, default=0.00)
    compare_at_price = models.DecimalField(_('Compare at price'), max_digits=10, decimal_places=2, default=0.00)
    price_varies = models.BooleanField(_('Price varies'), default=False)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    class Meta:
        verbose_name = _('Product variant')
        verbose_name_plural = _('Product variants')
        ordering = ['id']
        unique_together = ['product', 'slug']

    def __str__(self) -> str:
        return self.title

    @staticmethod
    def type() -> str:
        return 'variant'

    def get_absolute_url(self) -> str:
        try:
            return reverse('pcart_product:product-variant-detail', args=[self.product.slug, self.slug])
        except NoReverseMatch:
            return '#no-page-for-product-app'

    def get_properties_fields(self, initial_values: Optional[Dict] = None) -> List[Tuple]:
        return self.product.product_type.get_properties_fields(
            initial_values=self.properties if initial_values is None else initial_values,
            for_variants=True)

    def get_image(self) -> 'ProductImage':
        return self.product.get_image()

    def weight(self) -> float:
        return self.product.weight

    def available(self) -> bool:
        return self.price > 0 and self.status.show_buy_button


def generate_product_image_filename(instance: 'ProductImage', filename: str) -> str:
    ext = filename.split('.')[-1]
    url = 'images/products/%s/%s.%s' % (instance.product.id, str(uuid.uuid4()).replace('-', ''), ext)
    return url


class ProductImage(SortableMixin, TranslatableModel):
    """
    Represents a product image.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    translations = TranslatedFields(
        title=models.CharField(_('Title'), max_length=255, default='', blank=True)
    )

    product = SortableForeignKey(
        Product, verbose_name=_('Product'), related_name='images', on_delete=models.CASCADE)
    image = models.ImageField(_('Image'), null=True, blank=True, upload_to=generate_product_image_filename)
    html_snippet = models.TextField(_('HTML snippet'), default='', blank=True)

    tags = ArrayField(
        models.CharField(max_length=MAX_TAG_LENGTH),
        verbose_name=_('Tags'),
        blank=True,
        default=list,
    )

    download_link = models.CharField(_('Download link'), max_length=300, default='', blank=True)
    downloaded = models.BooleanField(_('Downloaded'), default=False)

    added = models.DateTimeField(_('Added'), auto_now_add=True)
    changed = models.DateTimeField(_('Changed'), auto_now=True)

    position = models.PositiveIntegerField(_('Position'), default=0, editable=False, db_index=True)

    class Meta:
        verbose_name = _('Product image')
        verbose_name_plural = _('Product images')
        ordering = ['position']

    def __str__(self) -> str:
        return self.title or self.product.title

    def download_product_image(self, save: bool = False):
        from django.core.files import File
        from ftplib import FTP
        from io import BytesIO
        import urllib.parse
        import os
        _changed = False
        o = urllib.parse.urlparse(self.download_link)

        if self.downloaded is False:
            if o.scheme in ['http', 'https']:
                # Download an image via HTTP
                local_filename, headers = urllib.request.urlretrieve(self.download_link)
                self.image.save(
                    os.path.basename(o.path),
                    File(open(local_filename, 'rb')),
                    save=False,
                )
                self.downloaded = True
                _changed = True
            elif o.scheme == 'ftp':
                # Download an image via FTP
                _ftp_sources = getattr(settings, 'PCART_FTP_IMAGE_SOURCES', {})
                _credentials = _ftp_sources.get(o.netloc)
                with FTP(o.netloc) as ftp:
                    if _credentials:
                        ftp.login(_credentials.get('user'), _credentials.get('password'))
                    else:
                        ftp.login()
                    ftp.getwelcome()
                    r = BytesIO()
                    ftp.retrbinary('RETR ~/%s' % o.path, r.write)
                    r.seek(0)

                    self.image.save(
                        os.path.basename(o.path),
                        File(r),
                        save=False,
                    )
                    self.downloaded = True
                    _changed = True

        if save and _changed:
            self.save()

    def save(self, *args, **kwargs):
        # Force image download if the link is specified
        if self.downloaded is False and self.download_link:
            self.download_product_image(save=False)
        super(ProductImage, self).save(*args, **kwargs)


@receiver(pre_save, sender=ProductImage)
def product_image_pre_save_listener(sender, instance: 'ProductImage', **kwargs):
    if instance.pk:
        try:
            p_im = ProductImage.objects.get(pk=instance.pk)
            if p_im.image and p_im.image != instance.image:
                p_im.image.delete(save=False)
        except ProductImage.DoesNotExist:
            pass


@receiver(post_delete, sender=ProductImage)
def product_image_post_delete_listener(sender, instance: 'ProductImage', **kwargs):
    if instance.image:
        instance.image.delete(save=False)


class PropertyValueSlugManager(models.Manager):
    """
    Queryset manager for `PropertyValueSlug` model. Filtering system is using it for autogeneration URL chunks
    for filtered colelction URLs.
    """
    def generate_unique_slug(self, value: str) -> str:
        """Returns a unique slug available for use for property value."""
        from pcart_core.utils import get_unique_slug
        slug = get_unique_slug(value, self.model)
        return slug

    def add_values(self, values: Optional[List] = None):
        values = [] if values is None else list(set(values))
        existing_values = self.filter(value__in=values).values_list('value', flat=True)
        for v in values:
            if v not in existing_values:
                slug = self.generate_unique_slug(v)
                item = self.model(value=v, slug=slug)
                item.save(using=self._db)

    def update_data(self, products=None):
        from .utils import get_unique_tagset, filter_non_numeric_tags
        values = filter_non_numeric_tags(get_unique_tagset(products))
        self.add_values(values)


class PropertyValueSlug(models.Model):
    """
    Represents property value slug. Such slugs are using by filtering system for making possible to autogenerate
    filtered collection URLs for property values which cannot be using as URL chunks without properly encoding.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    value = models.CharField(_('Value'), max_length=255, unique=True)
    slug = models.SlugField(_('Slug'), max_length=255, unique=True)

    objects = PropertyValueSlugManager()

    class Meta:
        verbose_name = _('Property value slug')
        verbose_name_plural = _('Property values slugs')
        ordering = ['value']

    def __str__(self) -> str:
        return self.value


# DjangoCMS plugins


class CollectionPluginModel(CMSPlugin):
    """ Represents a plugin with a products from the specified collection.
    """
    collection = TreeForeignKey(Collection, verbose_name=_('Collection'))
    show_title = models.BooleanField(_('Show title'), default=True)
    title = models.CharField(
        _('Title'), max_length=255, default='', blank=True,
        help_text=_('Optional. If not set the collection title will be used.')
    )
    filter_string = models.CharField(
        _('Filter string'), default='', blank=True, max_length=255,
        help_text=_('Use the list of filter chunks separated with / and ; characters.')
    )
    sorting = models.CharField(_('Sorting'), default='', blank=True, max_length=100)
    limit = models.PositiveIntegerField(_('Limit'), default=6)
    template_name = models.CharField(_('Template name'), default='', blank=True, max_length=100)
    data = JSONField(
        _('Data'), default=dict, blank=True,
        help_text=_('Optional. Some templates may support additional options for conditional rendering.'),
    )
    ajax_load = models.BooleanField(_('AJAX load'), default=False)

    def __init__(self, *args, **kwargs):
        super(CollectionPluginModel, self).__init__(*args, **kwargs)

    def __str__(self) -> str:
        return str(self.collection)


class SimilarProductsPluginModel(CMSPlugin):
    """
    Represents a plugin with a list of products with set of specified tags related to the product
    which page plugin inserted into.
    """
    PRICE_FILTERING_CHOICES = (
        ('', _('Do not filter')),
        ('absolute', _('Absolute difference')),
        ('relative', _('Relative difference')),
    )

    collection = TreeForeignKey(Collection, verbose_name=_('Collection'))
    show_title = models.BooleanField(_('Show title'), default=True)
    title = models.CharField(
        _('Title'), max_length=255, default='', blank=True,
        help_text=_('Optional. If not set the collection title will be used.')
    )
    tags_or_prefixes = ArrayField(
        models.CharField(max_length=MAX_TAG_LENGTH),
        verbose_name=_('Tags or prefixes'),
        blank=True,
        default=list,
        help_text=_('You can specify like full tag as `prefix:value` but also `prefix` only.')
    )
    vendor_filtering = models.BooleanField(_('Vendor filtering'), default=False)
    price_difference_filtering = models.CharField(
        _('Price difference filtering'), default='', blank=True,
        max_length=30,
        choices=PRICE_FILTERING_CHOICES)
    price_filtering_value = models.DecimalField(
        _('Price filtering value'), max_digits=10, decimal_places=2, default=0.00)

    sorting = models.CharField(_('Sorting'), default='', blank=True, max_length=100)
    limit = models.PositiveIntegerField(_('Limit'), default=6)
    template_name = models.CharField(_('Template name'), default='', blank=True, max_length=100)
    data = JSONField(
        _('Data'), default=dict, blank=True,
        help_text=_('Optional. Some templates may support additional options for conditional rendering.'),
    )
    ajax_load = models.BooleanField(_('AJAX load'), default=False)

    def __init__(self, *args, **kwargs):
        super(SimilarProductsPluginModel, self).__init__(*args, **kwargs)

    def __str__(self) -> str:
        return str(self.collection)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-29 14:42
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0049_remove_untranslated_fields_from_collection'),
    ]

    operations = [
        migrations.RenameField(
            model_name='collectiontranslation',
            old_name='description_t',
            new_name='description',
        ),
        migrations.RenameField(
            model_name='collectiontranslation',
            old_name='meta_description_t',
            new_name='meta_description',
        ),
        migrations.RenameField(
            model_name='collectiontranslation',
            old_name='meta_keywords_t',
            new_name='meta_keywords',
        ),
        migrations.RenameField(
            model_name='collectiontranslation',
            old_name='meta_text_t',
            new_name='meta_text',
        ),
        migrations.RenameField(
            model_name='collectiontranslation',
            old_name='page_title_t',
            new_name='page_title',
        ),
        migrations.RenameField(
            model_name='collectiontranslation',
            old_name='title_t',
            new_name='title',
        ),
    ]



from django.utils.translation import ugettext as _
from django.utils.safestring import mark_safe
from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from .models import CollectionPluginModel, SimilarProductsPluginModel
from .forms import CollectionPluginForm, SimilarProductsPluginForm


class CollectionPluginPublisher(CMSPluginBase):
    model = CollectionPluginModel  # model where plugin data are saved
    form = CollectionPluginForm
    module = _("Catalog")
    name = _("Collection")  # name of the plugin in the interface
    render_template = "catalog/plugins/collection_plugin.html"

    def render(self, context, instance, placeholder):
        from .filtering import ProductsFilterManager
        from .utils import filter_slug_to_tags

        if instance.ajax_load:
            filter_chunks = instance.filter_string
            if filter_chunks and not filter_chunks.startswith('/'):
                filter_chunks = '/' + filter_chunks
            _url = instance.collection.get_absolute_url() + filter_chunks
            if not _url.endswith('/'):
                _url += '/'
            _url += '?view=%s&sort=%s&limit=%s' % (
                instance.template_name, instance.sorting, instance.limit)
            _context = {
                'url': mark_safe(_url),
            }
            _context.update({'parent_context': context})
            _context.update({'instance': instance})
            return _context
        else:
            filter_manager = ProductsFilterManager(
                collection=instance.collection,
                view=instance.template_name,
                sort=instance.sorting,
                limit=instance.limit,
            )
            filter_chunks = instance.filter_string.split('/')
            filter_tags, vendors, prices, normalized_url_chunks, _redirect = filter_slug_to_tags(
                instance.collection, filter_chunks)
            filter_manager.set_filters(filter_tags=filter_tags, vendors=vendors, prices=prices)
            _context = filter_manager.get_context()
            _context.update({'parent_context': context})
            _context.update({'instance': instance})
            return _context


plugin_pool.register_plugin(CollectionPluginPublisher)


class SimilarProductsPluginPublisher(CMSPluginBase):
    model = SimilarProductsPluginModel  # model where plugin data are saved
    form = SimilarProductsPluginForm
    module = _("Catalog")
    name = _("Similar products")  # name of the plugin in the interface
    render_template = "catalog/plugins/similar_products_plugin.html"

    def render(self, context, instance, placeholder):
        from .filtering import ProductsFilterManager
        from .utils import filter_slug_to_tags, normalize_filter_tags

        product = context.get('product')
        _tags_for_filtering = []  # Here we will collect tags for filtering
        for t in instance.tags_or_prefixes:
            if ':' in t:
                # A tag found
                _tags_for_filtering.append(t)
            else:
                # A prefix only, so you should find a related tag for current product
                for i in product.tags:
                    _prefix, _value = i.split(':')
                    if t == _prefix:
                        _tags_for_filtering.append(i)
        _vendors = None
        if instance.vendor_filtering:
            _vendors = [product.vendor]
        _prices = None
        if instance.price_difference_filtering == 'absolute':
            _prices = (
                product.price - instance.price_filtering_value,
                product.price + instance.price_filtering_value,
            )
        elif instance.price_difference_filtering == 'relative':
            _prices = (
                float(int(product.price / instance.price_filtering_value))
                if instance.price_filtering_value != 0 else product.price,
                float(int(product.price * instance.price_filtering_value))
                if instance.price_filtering_value != 0 else product.price,
            )

        if instance.ajax_load:
            _url = instance.collection.get_absolute_url() + '/'.join(
                normalize_filter_tags(
                    instance.collection,
                    _vendors,
                    _prices,
                    _tags_for_filtering,
                )) + '/'
            _url += '?view=%s&sort=%s&limit=%s' % (
                instance.template_name, instance.sorting, instance.limit)
            _context = {
                'url': mark_safe(_url),
            }
            _context.update({'parent_context': context})
            _context.update({'instance': instance})
            return _context
        else:
            filter_manager = ProductsFilterManager(
                collection=instance.collection,
                view=instance.template_name,
                sort=instance.sorting,
                limit=instance.limit,
            )
            filter_manager.set_filters(
                filter_tags=_tags_for_filtering,
                vendors=_vendors,
                prices=_prices,
            )
            _context = filter_manager.get_context()
            _context.update({'parent_context': context})
            _context.update({'instance': instance})
            return _context


plugin_pool.register_plugin(SimilarProductsPluginPublisher)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-01 13:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0067_auto_20170930_2021'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionpluginmodel',
            name='ajax_load',
            field=models.BooleanField(default=False, verbose_name='AJAX load'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-29 13:38
from __future__ import unicode_literals

from django.db import migrations
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist


def forwards_func(apps, schema_editor):
    Vendor = apps.get_model('pcart_catalog', 'Vendor')
    VendorTranslation = apps.get_model('pcart_catalog', 'VendorTranslation')

    for obj in Vendor.objects.all():
        VendorTranslation.objects.create(
            master_id=obj.pk,
            language_code=settings.LANGUAGE_CODE,
            title_t=obj.title,
            description_t=obj.description,
        )


def backwards_func(apps, schema_editor):
    Vendor = apps.get_model('pcart_catalog', 'Vendor')
    VendorTranslation = apps.get_model('pcart_catalog', 'VendorTranslation')

    for obj in Vendor.objects.all():
        translation = _get_translation(obj, VendorTranslation)
        obj.title = translation.title_t
        obj.description = translation.description_t
        obj.save()   # Note this only calls Model.save()


def _get_translation(object, VendorTranslation):
    translations = VendorTranslation.objects.filter(master_id=object.pk)
    try:
        # Try default translation
        return translations.get(language_code=settings.LANGUAGE_CODE)
    except ObjectDoesNotExist:
        try:
            # Try default language
            return translations.get(language_code=settings.PARLER_DEFAULT_LANGUAGE_CODE)
        except ObjectDoesNotExist:
            # Maybe the object was translated only in a specific language?
            # Hope there is a single translation
            return translations.get()


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0043_add_translations_to_vendor'),
    ]

    operations = [
        migrations.RunPython(forwards_func, backwards_func),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-04-25 10:20
from __future__ import unicode_literals

from django.db import migrations, models


def prepopulate_vendor_slugs(apps, schema_editor):
    from pcart_core.utils import get_unique_slug
    Vendor = apps.get_model('pcart_catalog', 'Vendor')
    for vendor in Vendor.objects.all():
        vendor.slug = get_unique_slug(vendor.title, Vendor)
        vendor.save()


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0015_vendor_slug'),
    ]

    operations = [
        migrations.RunPython(prepopulate_vendor_slugs),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-29 18:22
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0061_migrate_translatable_fields_for_productimage'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='productimage',
            name='title',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 12:35
from __future__ import unicode_literals

import cms.models.fields
from django.db import migrations
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('pcart_catalog', '0022_auto_20170814_1258'),
    ]

    operations = [
        migrations.AddField(
            model_name='producttype',
            name='primary_placeholder',
            field=cms.models.fields.PlaceholderField(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='product_type_primary', slotname='primary_product_type_placeholder', to='cms.Placeholder'),
        ),
        migrations.AddField(
            model_name='producttype',
            name='secondary_placeholder',
            field=cms.models.fields.PlaceholderField(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='product_type_secondary', slotname='secondary_product_type_placeholder', to='cms.Placeholder'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 16:59
from __future__ import unicode_literals

import cms.models.fields
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('pcart_catalog', '0024_producttypeproperty_hidden'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='condition_max_price',
            field=models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Max price'),
        ),
        migrations.AddField(
            model_name='collection',
            name='condition_min_price',
            field=models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Min price'),
        ),
        migrations.AddField(
            model_name='collection',
            name='condition_product_types',
            field=models.ManyToManyField(blank=True, related_name='auto_collections', to='pcart_catalog.ProductType', verbose_name='Product types'),
        ),
        migrations.AddField(
            model_name='collection',
            name='condition_vendors',
            field=models.ManyToManyField(blank=True, related_name='auto_collections', to='pcart_catalog.Vendor', verbose_name='Vendors'),
        ),
        migrations.AddField(
            model_name='collection',
            name='exlude_tags',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, size=None, verbose_name='Exclude tags'),
        ),
        migrations.AddField(
            model_name='collection',
            name='extra_placeholder',
            field=cms.models.fields.PlaceholderField(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='collection_extra', slotname='secondary_collection_placeholder', to='cms.Placeholder'),
        ),
        migrations.AddField(
            model_name='collection',
            name='mandatory_tags',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, size=None, verbose_name='Mandatory tags'),
        ),
        migrations.AddField(
            model_name='collection',
            name='optional_tags',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, size=None, verbose_name='Optional tags'),
        ),
        migrations.AddField(
            model_name='collection',
            name='primary_placeholder',
            field=cms.models.fields.PlaceholderField(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='collection_primary', slotname='primary_collection_placeholder', to='cms.Placeholder'),
        ),
        migrations.AddField(
            model_name='collection',
            name='products_selection',
            field=models.CharField(default='manual', help_text='Use Automatic mode if you want to select products based on conditions.', max_length=30, verbose_name='Products selection'),
        ),
        migrations.AddField(
            model_name='collection',
            name='secondary_placeholder',
            field=cms.models.fields.PlaceholderField(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='collection_secondary', slotname='secondary_collection_placeholder', to='cms.Placeholder'),
        ),
    ]



from django.contrib import admin
from django.db import models
from mptt.admin import MPTTModelAdmin
from django.utils.translation import ugettext_lazy as _
from adminsortable.admin import NonSortableParentAdmin, SortableTabularInline, SortableAdmin
from django.utils.safestring import mark_safe
from cms.admin.placeholderadmin import FrontendEditableAdminMixin
from pcart_treeadmin.admin import TreeAdmin
from cms.admin.placeholderadmin import PlaceholderAdminMixin
from parler.admin import TranslatableAdmin, TranslatableTabularInline, TranslatableStackedInline
from .models import (
    Collection,
    ProductType,
    ProductTypeProperty,
    ProductStatus,
    Product,
    ProductVariant,
    ProductImage,
    Vendor,
    PropertyValueSlug,
)
from .forms import (
    EditProductVariantForm,
    EditProductVariantFormSet,
    EditCollectionForm,
    EditProductImageForm,
)
from .settings import PCART_ENABLE_PRODUCT_VARIANTS


class VendorAdmin(TranslatableAdmin):
    list_display = ('title', 'slug', 'has_image')
    search_fields = ['translations__title', 'slug', 'id']

    def get_prepopulated_fields(self, request, obj=None):
        return {"slug": ("title",)}

    def has_image(self, obj) -> bool:
        return bool(obj.image)
    has_image.short_description = _('Has image')
    has_image.boolean = True

    def get_queryset(self, request):
        qs = super(VendorAdmin, self).get_queryset(request)
        return qs.prefetch_related('translations').translated().order_by('translations__title')


admin.site.register(Vendor, VendorAdmin)


class CollectionAdmin(PlaceholderAdminMixin, TranslatableAdmin, TreeAdmin):
    form = EditCollectionForm
    list_display = ('title_preview', 'changed', 'published', 'site', 'slug')
    list_filter = ('published', 'site')
    preserve_filters = True
    search_fields = ['translations__title', 'id']
    fieldsets = (
        (None, {
            'fields': ('site', 'title', 'slug', 'parent', 'description', 'image'),
        }),
        (_('SEO'), {
            'fields': ('page_title', 'meta_description', 'meta_keywords', 'meta_text'),
            'classes': ('collapse',)
        }),
        (_('Product selection'), {
            'fields': (
                'products_selection',
                'condition_vendors',
                'condition_product_types',
                'condition_min_price',
                'condition_max_price',
                'mandatory_tags',
                'optional_tags',
            ),
            'classes': ('collapse',)
        }),
        (_('Filters'), {
            'fields': (
                'custom_url_filter_rules',
                'url_filter_rules',
                'show_vendor_filter', 'show_properties_filters', 'exclude_properties_filters',
                'show_price_filter',
            ),
            'classes': ('collapse',)
        }),
        (_('Publication'), {
            'fields': ('published',),
        }),
    )
    filter_horizontal = ('condition_vendors', 'condition_product_types')

    def get_prepopulated_fields(self, request, obj=None):
        return {"slug": ("title",)}

    def title_preview(self, obj):
        return mark_safe('%s%s' % ('&nbsp;'*obj.level*4, obj.title))
    title_preview.short_description = _('Title')
    # title_preview.admin_order_field = 'title'

    def get_queryset(self, request):
        qs = super(CollectionAdmin, self).get_queryset(request)
        return qs.prefetch_related('translations').translated()


admin.site.register(Collection, CollectionAdmin)


class ProductTypePropertyAdmin(SortableAdmin):
    list_display = (
        'title', 'product_type', 'default_value',
        'tag_prefix', 'use_in_filters', 'for_variants', 'hidden',
    )
    search_fields = ('title',)
    list_filter = ('product_type',)


admin.site.register(ProductTypeProperty, ProductTypePropertyAdmin)

# class ProductTypePropertyInline(SortableTabularInline):
#     model = ProductTypeProperty
#     exclude = tuple() if PCART_ENABLE_PRODUCT_VARIANTS else ('for_variants',)
#     extra = 1


class ProductTypeAdmin(PlaceholderAdminMixin, TranslatableAdmin):  # , NonSortableParentAdmin
    list_display = ('title', 'id')
    search_fields = ['translations__title']
    # inlines = [ProductTypePropertyInline]
    fieldsets = (
        (None, {
            'fields': ('title',),
        }),
        (_('SEO'), {
            'fields': ('page_title', 'meta_description', 'meta_keywords', 'meta_text'),
            'classes': ('collapse',),
        }),
        (_('Details'), {
            'fields': ('default_quantity_for_purchase',),
            'classes': ('collapse',),
        })
    )

    def get_queryset(self, request):
        qs = super(ProductTypeAdmin, self).get_queryset(request)
        return qs.prefetch_related('translations').translated()


admin.site.register(ProductType, ProductTypeAdmin)


class ProductStatusAdmin(TranslatableAdmin):
    list_display = ('title', 'show_buy_button', 'is_visible', 'is_searchable', 'weight')
    search_fields = ['translations__title']
    fieldsets = (
        (None, {
            'fields': ('title',),
        }),
        (_('Features'), {
            'fields': (
                'show_buy_button', 'is_visible', 'is_searchable',
            )
        }),
        (_('Priority'), {
            'fields': ('weight',),
        }),
    )

    def get_queryset(self, request):
        qs = super(ProductStatusAdmin, self).get_queryset(request)
        return qs.prefetch_related('translations').translated()


admin.site.register(ProductStatus, ProductStatusAdmin)


class ProductImageInline(SortableTabularInline, TranslatableTabularInline):
    model = ProductImage
    extra = 1
    form = EditProductImageForm


class ProductVariantInline(TranslatableTabularInline):
    model = ProductVariant
    readonly_fields = ['tags']
    form = EditProductVariantForm
    formset = EditProductVariantFormSet
    extra = 1

    def get_fields(self, request, obj=None):
        exclude_prop_fields = tuple(p[0] for p in obj.get_properties_fields(for_variants=True))
        self.exclude = exclude_prop_fields
        fields = ('title', 'slug', 'tags', 'sku', 'barcode', 'status', 'price', 'compare_at_price', 'price_varies') + exclude_prop_fields
        return fields


class ProductAdmin(FrontendEditableAdminMixin, TranslatableAdmin, NonSortableParentAdmin):
    list_display = (
        'show_title',
        # 'vendor',
        'status',
        'show_price',
        'product_type',
        'get_collections',
        'published')
    list_filter = ('published', 'status', 'product_type')
    filter_horizontal = ('collections',)
    preserve_filters = True
    search_fields = ['title', 'id']
    inlines = [ProductImageInline, ProductVariantInline]
    save_as_continue = True

    def get_prepopulated_fields(self, request, obj=None):
        return {"slug": ("title",)}

    def get_queryset(self, request):
        queryset = super(ProductAdmin, self).get_queryset(request)
        queryset = queryset \
            .prefetch_related('translations') \
            .prefetch_related('vendor')\
            .prefetch_related('vendor__translations')\
            .prefetch_related('status')\
            .prefetch_related('status__translations')\
            .prefetch_related('product_type')\
            .prefetch_related('product_type__translations')\
            .prefetch_related('collections')\
            .prefetch_related('collections__translations')
        return queryset

    def show_price(self, obj):
        if obj.variants_count > 0:
            return '%s - %s' % (obj.min_variant_price, obj.max_variant_price)
        else:
            if obj.compare_at_price != 0:
                return '%s --%s--' % (obj.price, obj.compare_at_price)
            else:
                return '%s' % obj.price
    show_price.short_description = _('Price')
    show_price.admin_order_field = 'min_variant_price'

    def get_collections(self, obj):
        return ', '.join([x.title for x in obj.collections.all()])
    get_collections.short_description = _('Collections')

    def show_title(self, obj):
        from django.utils.html import format_html
        if obj.variants_count > 0:
            return format_html(
                '{0} <span style="float:right;"><strong>({1})</strong></span>',
                obj.title,
                obj.variants_count)
        else:
            return obj.title
    show_title.short_description = _('Title')
    # show_title.admin_order_field = 'title'

    def get_form(self, request, obj=None, **kwargs):
        from .forms import AddProductForm, EditProductForm
        if obj is None:
            kwargs['form'] = AddProductForm
        else:
            kwargs['form'] = EditProductForm
        return super(ProductAdmin, self).get_form(request, obj, **kwargs)

    def get_fieldsets(self, request, obj=None):
        if obj is None:
            self.exclude = []
            return (
                (None, {
                    'fields': ('title', 'slug', 'product_type', 'vendor', 'status'),
                }),
            )
        else:
            exclude_prop_fields = tuple(p[0] for p in obj.get_properties_fields())
            self.exclude = exclude_prop_fields
            return (
                (None, {
                    'fields': (
                        'title', 'slug', 'external_id', 'product_type', 'vendor', 'description'),
                }),
                (_('SEO'), {
                    'fields': ('page_title', 'meta_description', 'meta_keywords', 'meta_text'),
                    'classes': ('collapse',),
                }),
                (_('Tags'), {
                    'fields': ('tags', 'sku', 'barcode', 'status', 'quantity'),
                }),
                (_('Properties'), {
                    'fields': exclude_prop_fields,
                }),
                (_('Price'), {
                    'fields': ('price', 'compare_at_price', 'price_varies', 'max_variant_price', 'min_variant_price'),
                }),
                (_('Shipping'), {
                    'fields': ('weight',),
                    'classes': ('collapse',),
                }),
                (_('Additional data'), {
                    'fields': (
                        'rank_1', 'rank_2', 'rank_3',
                        'extra_int_1', 'extra_int_2',
                        'extra_float_1', 'extra_float_2',
                        'extra_price_1', 'extra_price_2', 'extra_price_3',
                        'extra_text_1', 'extra_text_2',
                        'extra_data',
                        'extra_id_1', 'extra_id_2',
                    ),
                    'classes': ('collapse',),
                }),
                (_('Publication'), {
                    'fields': ('collections', 'published', 'variants_count', 'added', 'changed'),
                }),
            )

    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request, obj):
            # hide ProductVariantInline in the add view
            if isinstance(inline, ProductVariantInline) and obj is None:
                continue
            elif isinstance(inline, ProductVariantInline) and PCART_ENABLE_PRODUCT_VARIANTS is False:
                continue
            yield inline.get_formset(request, obj), inline

    def get_readonly_fields(self, request, obj=None):
        if obj is None:
            return []
        else:
            if obj.variants_count == 0:
                return (
                    'product_type',
                    'tags',
                    'min_variant_price',
                    'max_variant_price',
                    'variants_count',
                    'added',
                    'changed',
                )
            else:
                return (
                    'product_type',
                    'tags',
                    'status',
                    'min_variant_price',
                    'max_variant_price',
                    'variants_count',
                    'added',
                    'changed',
                )

admin.site.register(Product, ProductAdmin)


class PropertyValueSlugAdmin(admin.ModelAdmin):
    list_display = ('value', 'slug')
    search_fields = ('value', 'slug')

admin.site.register(PropertyValueSlug, PropertyValueSlugAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-28 09:51
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0032_auto_20170927_2141'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='extra_data',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, verbose_name='Extra data'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_float_1',
            field=models.FloatField(default=0.0, verbose_name='Extra float 1'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_float_2',
            field=models.FloatField(default=0.0, verbose_name='Extra float 2'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_id_1',
            field=models.CharField(blank=True, db_index=True, default='', max_length=255, verbose_name='Extra ID 1'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_id_2',
            field=models.CharField(blank=True, db_index=True, default='', max_length=255, verbose_name='Extra ID 2'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_int_1',
            field=models.IntegerField(default=0, verbose_name='Extra int 1'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_int_2',
            field=models.IntegerField(default=0, verbose_name='Extra int 2'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_price_1',
            field=models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Extra price 1'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_price_2',
            field=models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Extra price 2'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_price_3',
            field=models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Extra price 3'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_text_1',
            field=models.TextField(blank=True, default='', verbose_name='Extra text 1'),
        ),
        migrations.AddField(
            model_name='product',
            name='extra_text_2',
            field=models.TextField(blank=True, default='', verbose_name='Extra text 2'),
        ),
        migrations.AddField(
            model_name='product',
            name='rank_1',
            field=models.FloatField(default=0.0, verbose_name='Rank 1'),
        ),
        migrations.AddField(
            model_name='product',
            name='rank_2',
            field=models.FloatField(default=0.0, verbose_name='Rank 2'),
        ),
        migrations.AddField(
            model_name='product',
            name='rank_3',
            field=models.FloatField(default=0.0, verbose_name='Rank 3'),
        ),
        migrations.AddField(
            model_name='producttype',
            name='default_quantity_for_purchase',
            field=models.PositiveIntegerField(default=1, verbose_name='Default quantity for purchase'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-29 18:19
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0059_rename_translatable_fields_for_productvariant'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProductImageTranslation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language_code', models.CharField(db_index=True, max_length=15, verbose_name='Language')),
                ('title_t', models.CharField(blank=True, default='', max_length=255, verbose_name='Title')),
                ('master', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='pcart_catalog.ProductImage')),
            ],
            options={
                'db_tablespace': '',
                'db_table': 'pcart_catalog_productimage_translation',
                'verbose_name': 'Product image Translation',
                'default_permissions': (),
                'managed': True,
            },
        ),
        migrations.AlterUniqueTogether(
            name='productimagetranslation',
            unique_together=set([('language_code', 'master')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-14 09:58
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0021_collection_custom_url_filter_rules'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='price_varies',
            field=models.BooleanField(default=False, verbose_name='Price varies'),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='price_varies',
            field=models.BooleanField(default=False, verbose_name='Price varies'),
        ),
    ]



from django.urls import reverse, NoReverseMatch
from django import template
register = template.Library()


@register.filter
def filter_for_tags(queryset, filter_tags):
    return queryset.filter(tags__contains=filter_tags)


@register.filter
def published(queryset):
    return queryset.filter(published=True)


@register.simple_tag
def add_to_cart_url():
    try:
        return reverse('pcart_cart:add-to-cart')
    except NoReverseMatch:
        return '#no-page-for-cart'




def split_collection_slug(slug):
    slug_chunks = slug.split('/')
    collection_slug = slug_chunks[0]
    if len(slug_chunks) > 1:
        filter_chunks = slug_chunks[1:]
    else:
        filter_chunks = []
    return collection_slug, filter_chunks


def split_choices(line):
    pairs = line.split('|')
    result = {}
    for p in pairs:
        k, v = p.split('=', 1)
        if v:
            result.update({k: v})
    return result


def reverse_choices(line, prefix):
    choices = split_choices(line)
    result = dict([('%s:%s' % (prefix, a), b) for b, a in choices.items()])
    return result


def get_property_values_for_slugs():
    from .models import PropertyValueSlug
    return dict(PropertyValueSlug.objects.values_list('slug', 'value'))


def get_property_slugs():
    from .models import PropertyValueSlug
    return PropertyValueSlug.objects.values_list('value', 'slug')


def filter_slug_to_tags(collection, filter_chunks):
    from .models import Vendor
    import itertools
    _orig_filter_chunks = filter_chunks
    filter_tags = set()
    prices = (None, None)
    rules = collection.get_url_filter_rules()
    rules_table = [
        line.strip().split('/') for line in rules.splitlines() if not (line.startswith('#') or line.strip() == '')]

    excludes = []
    for rule in rules_table:
        if rule[0] == 'TAG' and rule[2] == 'CHOICE':
            choices = split_choices(rule[3])
            excludes += choices.keys()

    vendor_slugs = []
    filter_chunks = list(itertools.chain(*[c.split(';') for c in filter_chunks]))

    _allowed_vendor_slugs = {a: b for a, b in Vendor.objects.values_list('slug', 'id')}
    _check_price = False

    value_slugs = get_property_values_for_slugs()

    def get_value_for_slug(slug):
        return value_slugs.get(slug, slug)

    for chunk in filter_chunks:
        # Check by rule
        for rule in rules_table:
            if rule[0] == 'VENDOR' and chunk not in vendor_slugs and chunk in _allowed_vendor_slugs.keys():
                vendor_slugs.append(chunk)
            elif rule[0] == 'PRICE' and _check_price is False and chunk.startswith('price-'):
                _check_price = True
                _t = chunk.split('-')
                if len(_t) == 3:
                    try:
                        _from = float(_t[1])
                    except ValueError:
                        _from = None
                    try:
                        _to = float(_t[2])
                    except ValueError:
                        _to = None
                    prices = (_from, _to)
            elif rule[0] == 'TAG':
                cmd, prefix = rule[1][0], rule[1][1:]
                _tag = ''
                if cmd == '+' and chunk not in excludes and chunk.startswith(prefix):
                    if rule[2] == 'WITH PREFIX':
                        _tag = '%s:%s' % (prefix, get_value_for_slug(chunk))
                    elif rule[2] == 'VALUE ONLY':
                        _tag = '%s:%s' % (prefix, get_value_for_slug(chunk[len(prefix):]))
                    elif rule[2] == 'VALUE ONLY IGNORE DASH':
                        _tag = '%s:%s' % (prefix, get_value_for_slug(chunk[len(prefix)+1:]))
                    elif rule[2] == 'CHOICE':
                        choices = split_choices(rule[3])
                        if chunk in choices:
                            _tag = '%s:%s' % (prefix, get_value_for_slug(choices[chunk]))
                elif cmd == '-':
                    if rule[2] == 'CHOICE':
                        choices = split_choices(rule[3])
                        if chunk in choices:
                            _tag = '%s:%s' % (prefix, get_value_for_slug(choices[chunk]))
                if _tag:
                    filter_tags.add(_tag)
            elif rule[0] == 'NOT ALLOWED':
                choices = rule[1].split('|')
                if choices and set(choices).issubset(filter_tags):
                    if rule[2] == 'REMOVE':
                        filter_tags.remove(choices[int(rule[3])])

    vendors = Vendor.objects.filter(pk__in=[_allowed_vendor_slugs[i] for i in vendor_slugs]) if vendor_slugs else []

    filter_tags = list(filter_tags)
    normalized_f_tags = normalize_filter_tags(
        collection, vendors, prices,
        filter_tags, get_property_slugs(),
        rules,
    )
    _redirect = _orig_filter_chunks != normalized_f_tags
    return filter_tags, vendors, prices, normalized_f_tags, _redirect


def group_tags_for_prefixes(tags):
    prefixes_and_tags = [t.split(':') for t in tags]
    prefixes = list(set([t[0] for t in prefixes_and_tags]))
    regrouped_tags = [
        [':'.join(t) for t in prefixes_and_tags if t[0] == x] for x in prefixes]
    return regrouped_tags


def regroup_tags_for_filter_sets(tags):
    import itertools
    regrouped_tags = group_tags_for_prefixes(tags)
    result = list(itertools.product(*regrouped_tags))
    return result


def get_tags_array_query_expression(tag_groups, fieldname):
    from django.db.models import Q
    expr = None
    for g in tag_groups:
        kwargs = {'%s__contains' % fieldname: list(g)}
        if expr is None:
            expr = Q(**kwargs)
        else:
            expr |= Q(**kwargs)
    return expr


def generate_tags_for_product(product):
    properties = product.product_type.properties.filter(use_in_filters=True)
    tags = []
    for p in properties:
        property_value = product.properties.get(p.title)
        if property_value is not None:
            tags.append(p.tag_prefix + property_value)
    return tags


def normalize_filter_tags(
        collection, vendors, prices=(None, None), tags=None, cached_value_slugs=None,
        cached_url_filter_rules=None):
    if tags is None:
        tags = []
    if cached_value_slugs is None:
        cached_value_slugs = []
    result = []

    rules = cached_url_filter_rules or collection.get_url_filter_rules()
    rules_table = [
        line.strip().split('/') for line in rules.splitlines() if not (line.startswith('#') or line.strip() == '')]

    value_slugs = dict(cached_value_slugs)

    def get_slug_for_value(value):
        return value_slugs.get(value, value)

    for rule in rules_table:
        if rule[0] == 'VENDOR' and vendors:
            _vendor_slugs = []
            for v in vendors:
                if type(v) is str:
                    _vendor_slugs.append(v)
                else:
                    _vendor_slugs.append(v.slug)
            result.append(';'.join(_vendor_slugs))
        elif rule[0] == 'PRICE' and (bool(prices[0]) or bool(prices[1])):
            result.append('price-%s-%s' % (prices[0] or 'none', prices[1] or 'none'))
        elif rule[0] == 'TAG':
            cmd, prefix = rule[1][0], rule[1][1:]
            _tags = sorted([t for t in tags if t.split(':')[0] == prefix])
            _chunk = ''

            if cmd == '+':
                if rule[2] == 'WITH PREFIX':
                    _chunk = ';'.join(['%s%s' % (prefix, get_slug_for_value(t.replace(':', ''))) for t in _tags])
                elif rule[2] == 'VALUE ONLY':
                    _chunk = ';'.join(['%s%s' % (prefix, get_slug_for_value(t[len(prefix)+1:])) for t in _tags])
                elif rule[2] == 'VALUE ONLY IGNORE DASH':
                    _chunk = ';'.join(['%s-%s' % (prefix, get_slug_for_value(t[len(prefix)+1:])) for t in _tags])
                elif rule[2] == 'CHOICE':
                    choices = reverse_choices(rule[3])
                    for c in choices:
                        if c in _tags:
                            _chunk = '%s%s' % (prefix, get_slug_for_value(choices[c]))
            elif cmd == '-':
                if rule[2] == 'CHOICE':
                    choices = reverse_choices(rule[3], prefix)
                    for c in choices:
                        if c in _tags:
                            _chunk = get_slug_for_value(choices[c])
            if _chunk:
                result.append(_chunk)

    # print('normalize_filter_tags', collection, vendors, prices, tags, '->', result)

    return result


def get_unique_tagset(products=None):
    """
    Returns the set of tags for the specified products or for all products.

    :param products: A Product queryset. If None function is working with
       all product base.
    :return: A set of tags.
    """
    import itertools
    if products is None:
        from .models import Product
        products = Product.objects.all()

    tagset = products.values_list('tags', flat=True)
    tagset = list(set(itertools.chain(*tagset)))
    return tagset


def filter_non_numeric_tags(tagset):
    """
    Returns a list of non numeric tags without prefixes for specified tagset.

    :param tagset: A set of tags.
    :return: A filtered list of tags without prefixes.
    """
    values = list(
        filter(
            lambda x: not x.isnumeric(),
            map(
                lambda x: x.split(':', 1)[-1],
                tagset,
            ),
        ))
    return values



from django.contrib.sitemaps import GenericSitemap
from .models import Collection, Product, ProductVariant


COLLECTIONS_SITEMAP_CONFIG = {
    'queryset': Collection.objects.filter(published=True),
    'date_field': 'changed',
}


CollectionSitemap = GenericSitemap(COLLECTIONS_SITEMAP_CONFIG)


PRODUCT_SITEMAP_CONFIG = {
    'queryset': Product.objects.filter(published=True).prefetch_related('images'),
    'date_field': 'changed',
}


ProductSitemap = GenericSitemap(PRODUCT_SITEMAP_CONFIG, changefreq='daily')


PRODUCT_VARIANT_SITEMAP_CONFIG = {
    'queryset': ProductVariant.objects.filter(product__published=True),
    'date_field': 'changed',
}


ProductVariantSitemap = GenericSitemap(PRODUCT_VARIANT_SITEMAP_CONFIG, changefreq='daily')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-30 17:21
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0066_collectionpluginmodel_limit'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionpluginmodel',
            name='data',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Optional. Some templates may support additional options for conditional rendering.', verbose_name='Data'),
        ),
        migrations.AddField(
            model_name='collectionpluginmodel',
            name='show_title',
            field=models.BooleanField(default=True, verbose_name='Show title'),
        ),
        migrations.AddField(
            model_name='collectionpluginmodel',
            name='title',
            field=models.CharField(blank=True, default='', help_text='Optional. If not set the collection title will be used.', max_length=255, verbose_name='Title'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 18:12
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0028_auto_20170927_2047'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='meta_keywords',
            field=models.TextField(blank=True, verbose_name='Meta keywords'),
        ),
        migrations.AddField(
            model_name='collection',
            name='meta_text',
            field=models.TextField(blank=True, verbose_name='Meta text'),
        ),
        migrations.AddField(
            model_name='product',
            name='meta_keywords',
            field=models.TextField(blank=True, verbose_name='Meta keywords'),
        ),
        migrations.AddField(
            model_name='product',
            name='meta_text',
            field=models.TextField(blank=True, verbose_name='Meta text'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 17:06
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0026_auto_20170927_2003'),
    ]

    operations = [
        migrations.RenameField(
            model_name='collection',
            old_name='exlude_tags',
            new_name='exclude_tags',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-30 09:47
from __future__ import unicode_literals

from django.db import migrations
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0063_rename_translatable_fields_for_productimage'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectiontranslation',
            name='description',
            field=djangocms_text_ckeditor.fields.HTMLField(blank=True, verbose_name='Description'),
        ),
    ]



from django.template.loader import render_to_string
from django.conf import settings
from django.db.models import (
    Q, Prefetch, Count, Sum, Case,
    Min, Max, When,
    PositiveIntegerField, DecimalField, F)
from .models import (
    ProductType,
    ProductTypeProperty,
    Product,
    ProductVariant,
    ProductStatus,
    Vendor,
    PropertyValueSlug,
)
from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage
from .utils import (
    regroup_tags_for_filter_sets,
    get_tags_array_query_expression,
    normalize_filter_tags,
    group_tags_for_prefixes,
)
from . import settings as pcart_settings
from typing import Dict, List, Optional


class ProductsFilterManager:
    MAX_PAGINATION_LIMIT = 50

    def __init__(
            self, collection,
            view='default',
            sort=pcart_settings.PCART_COLLECTION_DEFAULT_ORDERING,
            limit=None):
        self.collection = collection
        self.url_filter_rules = collection.get_url_filter_rules()

        # Initial view
        self.collection_view = view
        self.template_name = self.get_view_attribute('template')
        self.exclude_filters_info = self.get_view_attribute('exclude_filters_info', False)

        if limit is not None and limit > self.MAX_PAGINATION_LIMIT:
            limit = self.MAX_PAGINATION_LIMIT
        self.paginate_by = limit or self.get_view_attribute('paginate_by')

        self.ordering_type = sort
        self.page = 1

        # Initial filtering
        self.filter_tags = []
        self._filtered = False

        self.vendors = Vendor.objects.none()
        self.prices = (None, None)
        self._prices = False

        # Init product types
        self.product_types = self.get_product_types()
        self.properties_for_filters = self.get_properties_for_filters()
        self.visible_status_ids = self.get_visible_status_ids()
        self.property_value_slugs = self.get_property_value_slugs()

        self._products_kwargs_all = self._products_kwargs_without_variants = self._variants_kwargs = dict()

    def set_page(self, page: int = 1):
        self.page = page

    def set_sort(self, sort: str = pcart_settings.PCART_COLLECTION_DEFAULT_ORDERING):
        self.ordering_type = sort

    def get_view_settings(self) -> Dict:
        return settings.PCART_COLLECTION_TEMPLATES[self.collection_view]

    def get_view_attribute(self, attribute, default=None):
        return self.get_view_settings().get(attribute, default)

    def set_filters(self, filter_tags=None, vendors=None, prices=None):
        if filter_tags is not None:
            self.filter_tags = filter_tags
        if vendors is not None:
            self.vendors = vendors
        if prices is not None:
            self.prices = prices
            if self.prices[0] or self.prices[1]:
                self._prices = True
            else:
                self._prices = False

    def set_view(self, view: str = 'default'):
        self.collection_view = view
        self.template_name = self.get_view_attribute('template')
        self.exclude_filters_info = self.get_view_attribute('exclude_filters_info', False)
        self.paginate_by = self.get_view_attribute('paginate_by')

    def get_product_types(self):
        return self.collection.get_product_types()

    def get_properties_for_filters(self):
        result = ProductTypeProperty.objects.filter(
            product_type__in=self.product_types, use_in_filters=True) \
            .order_by('tag_prefix') \
            .values_list('title', 'tag_prefix', 'for_variants') \
            .distinct('tag_prefix')
        return result

    def get_visible_status_ids(self):
        return ProductStatus.objects.filter(is_visible=True).values_list('id', flat=True)

    def get_property_value_slugs(self):
        return PropertyValueSlug.objects.values_list('value', 'slug')

    def get_products_kwargs_all(self) -> Dict:
        if self.collection.products_selection == 'manual':
            return {
                'collections': self.collection,
                'published': True,
                'status_id__in': self.visible_status_ids,
            }
        elif self.collection.products_selection == 'automatic':
            _result = {
                'published': True,
                'status_id__in': self.visible_status_ids,
            }
            if self.collection.condition_vendors.exists():
                _result.update({'vendor__in': self.collection.condition_vendors.all()})
            if self.collection.condition_product_types.exists():
                _result.update({'product_type__in': self.collection.condition_product_types.all()})
            if self.collection.condition_min_price > 0.0:
                _result.update({'price__gte': self.collection.condition_min_price})
            if self.collection.condition_max_price > 0.0:
                _result.update({'price__lte': self.collection.condition_max_price})
            if self.collection.mandatory_tags:
                _result.update({'tags__contains': self.collection.mandatory_tags})
            if self.collection.optional_tags:
                _result.update({'tags__overlap': self.collection.optional_tags})
            return _result

    def get_products_kwargs_without_variants(self):
        return dict(**self.get_products_kwargs_all(), **{'variants_count': 0})

    def get_variants_kwargs(self):
        return {
            'product__collections': self.collection,
            'product__published': True,
            'status_id__in': self.visible_status_ids,
        }

    def get_products(self):
        tag_groups = regroup_tags_for_filter_sets(self.filter_tags)
        tags_expr = get_tags_array_query_expression(tag_groups, 'tags')
        v_tags_expr = tags_expr
        variants__tags_expr = get_tags_array_query_expression(tag_groups, 'variants__tags')

        self._products_kwargs_all = self.get_products_kwargs_all()
        self._products_kwargs_without_variants = self.get_products_kwargs_without_variants()
        self._variants_kwargs = self.get_variants_kwargs()

        products = Product.objects.filter(**self._products_kwargs_all)
        total_max_price = products.aggregate(Max('max_variant_price'))['max_variant_price__max']

        self._filtered = False
        tag_expr_without_prices = tags_expr
        v_tag_expr_without_prices = v_tags_expr

        # Filter by price
        if self._prices:
            if self.prices[0] and self.prices[1]:
                p_price_expr_ex = \
                    Q(variants_count=0) & (Q(price__lt=self.prices[0]) | Q(price__gt=self.prices[1]))
                v_price_expr = Q(price__gte=self.prices[0]) & Q(price__lte=self.prices[1])
                variants_price_expr = Q(variants__price__gte=self.prices[0]) & Q(variants__price__lte=self.prices[1])
            elif self.prices[0] and self.prices[1] is None:
                p_price_expr_ex = Q(variants_count=0) & Q(price__lt=self.prices[0])
                v_price_expr = Q(price__gte=self.prices[0])
                variants_price_expr = Q(variants__price__gte=self.prices[0])
            elif self.prices[0] is None and self.prices[1]:
                p_price_expr_ex = Q(variants_count=0) & Q(price__gt=self.prices[1])
                v_price_expr = Q(price__lte=self.prices[1])
                variants_price_expr = Q(variants__price__lte=self.prices[1])

            self._filtered = True
            tags_expr = tags_expr & ~p_price_expr_ex
            v_tags_expr = v_tags_expr & v_price_expr
            variants__tags_expr = variants__tags_expr & variants_price_expr

        # Filter by Vendor
        if self.vendors:
            self._filtered = True
            products = products.filter(vendor__in=self.vendors)

        if self.filter_tags:
            self._filtered = True

        # Prefetch variants
        if self._filtered:
            products = products.filter(tags_expr)
            prefetch = Prefetch(
                'variants', queryset=ProductVariant.objects.filter(v_tags_expr))
        else:
            prefetch = Prefetch('variants')

        # Prefetch additional data
        products = products.prefetch_related(prefetch).prefetch_related('variants__status')
        products = products.prefetch_related('translations')
        products = products.prefetch_related('status')
        products = products.prefetch_related('status__translations')
        products = products.prefetch_related('images')
        products = products.prefetch_related('images__translations')
        products = products.prefetch_related('collections')
        products = products.prefetch_related('collections__translations')
        products = products.prefetch_related('product_type')
        products = products.prefetch_related('product_type__translations')

        # Add count, min and max annotations
        if self._filtered:
            products = products.annotate(
                filtered_variants_count=Sum(Case(
                    When(variants__tags_expr, then=1),
                    default=0,
                    output_field=PositiveIntegerField(),
                )),
                filtered_min_variant_price=Min(Case(
                    When(variants__tags_expr, then=F('variants__price')),
                    default=F('max_variant_price'),
                    output_field=DecimalField(),
                )),
                filtered_max_variant_price=Max(Case(
                    When(variants__tags_expr, then=F('variants__price')),
                    default=F('min_variant_price'),
                    output_field=DecimalField(),
                )),
            )
            products = products.exclude(Q(filtered_variants_count=0) & Q(variants_count__gt=0))

        if not self.exclude_filters_info:
            all_tags, vendor_tags = self.calculate_counters(tags_expr, v_tags_expr)
        else:
            all_tags = vendor_tags = list()

        ordering = self.get_ordering()
        products = products.order_by(*ordering)

        price_data = self.aggregate_price_data(
            products, total_max_price, tag_expr_without_prices, v_tag_expr_without_prices)

        return products, price_data, all_tags, vendor_tags

    def get_ordering(self):
        return pcart_settings.PCART_COLLECTION_ORDERINGS[
            self.ordering_type]['with_filters' if self._filtered else 'without_filters']

    def calculate_counters(self, tags_expr, v_tags_expr):
        import itertools
        from collections import Counter

        nonfiltered_vendor_counters = {
            x['vendor__slug']: x['id__count']
            for x in Product.objects.filter(
                **self._products_kwargs_without_variants).values('vendor__slug').annotate(Count('id')).order_by()}
        nonfiltered_v_vendor_counters = {
            x['product__vendor__slug']: x['id__count']
            for x in ProductVariant.objects.filter(
                **self._variants_kwargs).values('product__vendor__slug').annotate(Count('id')).order_by()}
        filtered_vendor_counters = {
            x['vendor__slug']: x['id__count']
            for x in Product.objects.filter(
                tags_expr,
                **self._products_kwargs_without_variants).values('vendor__slug').annotate(Count('id')).order_by()}
        filtered_v_vendor_counters = {
            x['product__vendor__slug']: x['id__count']
            for x in ProductVariant.objects.filter(
                v_tags_expr,
                **self._variants_kwargs).values('product__vendor__slug').annotate(Count('id')).order_by()}

        vendor_counter = Counter(filtered_vendor_counters)
        del vendor_counter[None]
        v_vendor_counter = Counter(filtered_v_vendor_counters)
        del v_vendor_counter[None]
        t_vendor_counter = vendor_counter + v_vendor_counter

        nonfiltered_vendor_counter = Counter(nonfiltered_vendor_counters)
        del nonfiltered_vendor_counter[None]
        nonfiltered_v_vendor_counter = Counter(nonfiltered_v_vendor_counters)
        del nonfiltered_v_vendor_counter[None]
        nonfiltered_t_vendor_counter = nonfiltered_vendor_counter + nonfiltered_v_vendor_counter

        _vendors_slugs = [v.slug for v in self.vendors]
        _vendors_titles = {
            slug: title for slug, title in Vendor.objects.values_list('slug', 'translations__title').distinct('slug')}

        vendor_tags = sorted([{
            'tag': t,
            'label': _vendors_titles.get(t),
            'count': nonfiltered_t_vendor_counter[t],
            'filtered': t_vendor_counter[t],
            'selected': t in _vendors_slugs,
            'active_filter': len(_vendors_slugs) > 0,
            'url': '' if t_vendor_counter[t] == 0 else self.collection.get_absolute_url() + '/'.join(
                normalize_filter_tags(
                    self.collection,
                    (
                        list(self.vendors) + [t] if t is not None else []
                    ) if t not in _vendors_slugs else [v.slug for v in self.vendors if v.slug != t],
                    self.prices,
                    self.filter_tags,
                    self.property_value_slugs,
                    self.url_filter_rules,
                )),
        } for t in nonfiltered_t_vendor_counter], key=lambda x: x['label'])

        all_tags = []

        ft_groups = group_tags_for_prefixes(self.filter_tags)
        filter_prefixes = set([i.split(':')[0] for i in self.filter_tags])

        _nonfiltered_tagset = list(
            Product.objects.filter(
                **self._products_kwargs_without_variants).values_list('tags', flat=True))

        _nonfiltered_v_tagset = list(
            ProductVariant.objects.filter(**self._variants_kwargs).values_list('tags', flat=True))

        for f in self.properties_for_filters:
            tags_chunk = [k for k in self.filter_tags if k.split(':')[0] != f[1]]
            _prefix = f[1]

            tag_groups = regroup_tags_for_filter_sets(tags_chunk)
            tags_expr = get_tags_array_query_expression(tag_groups, 'tags')

            _args = [tags_expr]
            _kwargs = dict(self._products_kwargs_without_variants)
            if self.vendors:
                _kwargs.update({'vendor__in': self.vendors})

            tagset = list(Product.objects.filter(*_args, **_kwargs).values_list('tags', flat=True))
            tagset = [x for x in itertools.chain(*[t for t in tagset]) if x.startswith(_prefix+':')]
            nonfiltered_tagset = [
                x for x in itertools.chain(*[t for t in _nonfiltered_tagset]) if x.startswith(_prefix + ':')]

            _args = [tags_expr]
            _kwargs = dict(self._variants_kwargs)
            if self.vendors:
                _kwargs.update({'product__vendor__in': self.vendors})

            v_tagset = list(ProductVariant.objects.filter(*_args, **_kwargs).values_list('tags', flat=True))
            v_tagset = [x for x in itertools.chain(*[t for t in v_tagset]) if x.startswith(_prefix+':')]
            nonfiltered_v_tagset = [
                x for x in itertools.chain(*[t for t in _nonfiltered_v_tagset]) if x.startswith(_prefix + ':')]

            counter = Counter(tagset)
            v_counter = Counter(v_tagset)
            t_counter = counter + v_counter

            nonfiltered_counter = Counter(nonfiltered_tagset)
            nonfiltered_v_counter = Counter(nonfiltered_v_tagset)
            nonfiltered_t_counter = nonfiltered_counter + nonfiltered_v_counter

            all_tags.append({
                'group_label': f[0],
                'tags': sorted([
                {
                    'tag': t,
                    'label': t.split(':')[-1],
                    'prefix': _prefix,
                    'selected': t in self.filter_tags,
                    'active_filter': _prefix in filter_prefixes,
                    'count': nonfiltered_t_counter[t],
                    'url': '' if t_counter[t] == 0 else self.collection.get_absolute_url() + '/'.join(
                        normalize_filter_tags(
                            self.collection,
                            self.vendors,
                            self.prices,
                            (
                                self.filter_tags+[t]
                            ) if t not in self.filter_tags else [x for x in self.filter_tags if x != t],
                            self.property_value_slugs,
                            self.url_filter_rules,
                        )),
                    'filtered': t_counter[t],
                } for t in nonfiltered_t_counter], key=self.tags_sort_func)})

        return all_tags, vendor_tags

    @staticmethod
    def tags_sort_func(x):
        import re

        def atoi(text):
            return int(text) if text.isdigit() else text

        def natural_keys(text):
            '''
            alist.sort(key=natural_keys) sorts in human order
            http://nedbatchelder.com/blog/200712/human_sorting.html
            (See Toothy's implementation in the comments)
            '''
            return [atoi(c) for c in re.split(r'(\d+)', text)]


        value = x['label'].strip()
        t = value.replace(',', '.').replace('/', '.').replace(' ', '')
        value = natural_keys(t)
        return value

    def aggregate_price_data(self, products, total_max_price, tag_expr_without_prices, v_tag_expr_without_prices):
        if self._filtered:
            p_agg_data = products.filter(variants_count=0).aggregate(Max('price'), Min('price'))
            v_agg_data = products.filter(variants_count__gt=0).aggregate(
                Min('filtered_min_variant_price'),
                Max('filtered_max_variant_price'),
                Sum('filtered_variants_count'),
            )
            variants_count = v_agg_data['filtered_variants_count__sum']
            min_price = min(
                [x for x in [p_agg_data['price__min'], v_agg_data['filtered_min_variant_price__min']] if
                 x is not None] or [0])
            max_price = max(
                [x for x in [p_agg_data['price__max'], v_agg_data['filtered_max_variant_price__max']] if
                 x is not None] or [total_max_price])

            pa_query = Product.objects.filter(tag_expr_without_prices, **self._products_kwargs_without_variants)
            va_query = ProductVariant.objects.filter(
                v_tag_expr_without_prices,
                **self._variants_kwargs)
            if self.vendors:
                pa_query = pa_query.filter(vendor__in=self.vendors)
                va_query = va_query.filter(product__vendor__in=self.vendors)
            pa_agg_data = pa_query.aggregate(Max('price'), Min('price'))
            va_agg_data = va_query.aggregate(Max('price'), Min('price'))

            min_available_price = min(
                [x for x in [pa_agg_data['price__min'], va_agg_data['price__min']] if x is not None] or [0])
            max_available_price = max(
                [x for x in [pa_agg_data['price__max'], va_agg_data['price__max']] if x is not None] or [
                    total_max_price])
        else:
            p_agg_data = products.filter(variants_count=0).aggregate(Max('price'), Min('price'))
            v_agg_data = products.filter(variants_count__gt=0).aggregate(
                Min('min_variant_price'),
                Max('max_variant_price'),
                Sum('variants_count'),
            )
            variants_count = v_agg_data['variants_count__sum']
            min_price = min(
                [x for x in [p_agg_data['price__min'], v_agg_data['min_variant_price__min']] if x is not None] or [0])
            max_price = max(
                [x for x in [p_agg_data['price__max'], v_agg_data['max_variant_price__max']] if x is not None] or [
                    total_max_price])

            pa_query = Product.objects.filter(**self._products_kwargs_without_variants)
            va_query = ProductVariant.objects.filter(**self._variants_kwargs)
            pa_agg_data = pa_query.aggregate(Max('price'), Min('price'))
            va_agg_data = va_query.aggregate(Max('price'), Min('price'))

            min_available_price = min(
                [x for x in [pa_agg_data['price__min'], va_agg_data['price__min']] if x is not None] or [0])
            max_available_price = max(
                [x for x in [pa_agg_data['price__max'], va_agg_data['price__max']] if x is not None] or [
                    total_max_price])

        data = {
            'variants_count': variants_count,
            'prices': self.prices,
            'min_available_price': min_available_price,
            'max_available_price': max_available_price,
            'min_price': min_price,
            'max_price': max_price,
            'min_selected_price': self.prices[0] or min_price,
            'max_selected_price': self.prices[1] or max_price,
        }
        return data

    def paginate_products(self, products):
        _products_list = products
        if self.paginate_by is None and _products_list.count() > self.MAX_PAGINATION_LIMIT:
            self.paginate_by = self.MAX_PAGINATION_LIMIT

        if self.paginate_by is not None:
            paginator = Paginator(_products_list, self.paginate_by)
            try:
                products = paginator.page(self.page)
            except PageNotAnInteger:
                products = paginator.page(1)
            except EmptyPage:
                products = paginator.page(paginator.num_pages)
        return products

    def get_context(self):
        products, price_data, all_tags, vendor_tags = self.get_products()
        context = {
            'collection': self.collection,
            'products': self.paginate_products(products),
            'filter_tags': self.filter_tags,
            'all_tags': all_tags,
            'vendor_tags': vendor_tags,
            'ordering_type': self.ordering_type,
            'filtered': self._filtered,
            'page_number': self.page,
            'collection_view': self.collection_view,
            'template_name': self.template_name,
        }
        context.update(price_data)
        return context

    def render_to_string(self, request=None):
        return render_to_string(self.template_name, context=self.get_context(), request=request)



# from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.http import Http404, QueryDict, HttpResponseNotFound, JsonResponse, HttpResponseForbidden
from django.views.decorators.http import require_http_methods
from django.shortcuts import redirect, render
from django.contrib.sites.shortcuts import get_current_site
from django.views.decorators.csrf import csrf_exempt
from django.core.cache import cache
from django.conf import settings
from . import settings as pcart_settings
from .models import (
    Collection,
    Product,
    ProductVariant,
)
from .utils import (
    split_collection_slug,
    filter_slug_to_tags,
    normalize_filter_tags,
)


AVAILABLE_COLLECTION_PARAMS = ['page', 'sort', 'view', 'json', 'limit']
COLLECTION_CACHE_TIMEOUT = 60 * 15


@require_http_methods(['GET'])
def collection_view(request, slug, view='default'):
    from .filtering import ProductsFilterManager

    is_ajax = request.is_ajax()

    _params = QueryDict(mutable=True)
    for k, v in request.GET.items():
        if k in AVAILABLE_COLLECTION_PARAMS:
            _params.update({k: v})

    json_response = request.GET.get('json', False) == 'true'
    ordering_type = request.GET.get('sort', pcart_settings.PCART_COLLECTION_DEFAULT_ORDERING)
    collection_view = request.GET.get('view', view)

    collection_view_config = getattr(settings, 'PCART_COLLECTION_TEMPLATES', {}).get(collection_view)
    if collection_view_config is None:
        # Unknown view
        raise Http404
    else:
        if collection_view_config.get('ajax_only', False) and not is_ajax:
            # Ajax only view cannot be accessible via regular request
            return HttpResponseForbidden('Invalid request.')

    limit = request.GET.get('limit')
    if limit is not None:
        try:
            limit = int(limit)
        except ValueError:
            limit = None

    page = request.GET.get('page', 1)
    cache_key = '%s?%s%s' % (
        request.path_info,
        _params.urlencode(),
        '|ajax' if is_ajax else '')
    result = cache.get(cache_key)
    if result and not json_response:
        return result

    collection_slug, filter_chunks = split_collection_slug(slug)
    try:
        collection = Collection.objects.get(slug=collection_slug, site_id=get_current_site(request).id)
    except Collection.DoesNotExist:
        response = HttpResponseNotFound('collection not found')
        cache.set(cache_key, response, COLLECTION_CACHE_TIMEOUT)
        return response

    filter_tags, vendors, prices, normalized_url_chunks, _redirect = filter_slug_to_tags(collection, filter_chunks)
    if _redirect:
        redirect_url = collection.get_absolute_url() + '/'.join(normalized_url_chunks)
        if not redirect_url.endswith('/'):
            redirect_url += '/'
        response = redirect(redirect_url)
        cache.set(cache_key, response, COLLECTION_CACHE_TIMEOUT)
        return response

    filter_manager = ProductsFilterManager(
        collection=collection, view=collection_view, sort=ordering_type, limit=limit)
    filter_manager.set_page(page)
    filter_manager.set_filters(filter_tags=filter_tags, vendors=vendors, prices=prices)

    context = filter_manager.get_context()
    context.update({
        'is_ajax': is_ajax,
        'page_url': request.path,
    })
    if json_response:
        context['collection'] = collection.as_dict()
        context['products'] = list(map(lambda x: x.as_dict(), context['products']))
        response = JsonResponse(context, safe=True)
        return response
    else:
        response = render(request, context['template_name'], context)
        cache.set(cache_key, response, COLLECTION_CACHE_TIMEOUT)
        return response



@csrf_exempt
@require_http_methods(['POST'])
def filter_form_dispatcher(request, collection_slug):
    vendors = request.POST.getlist('vendor')
    filter_tags = request.POST.getlist('tag')

    price = request.POST.get('price')
    if price is not None:
        price_from, price_to = price.split(request.POST.get('price-delimiter', ','))
    else:
        price_from = request.POST.get('price-from')
        price_to = request.POST.get('price-to')

    try:
        price_from = float(price_from)
    except ValueError:
        price_from = None
    try:
        price_to = float(price_to)
    except ValueError:
        price_to = None
    prices = (price_from, price_to)

    collection = Collection.objects.get(slug=collection_slug)
    _url = collection.get_absolute_url() + '/'.join(
        normalize_filter_tags(
            collection,
            vendors,
            prices,
            filter_tags,
        )) + '/'
    return redirect(_url)


def redirect_to_collections(request):
    return redirect('pcart_collection:all-collections')


class ProductDetailView(DetailView):
    model = Product
    context_object_name = 'product'
    template_name = 'catalog/product.html'

    def get_context_data(self, **kwargs):
        context = super(ProductDetailView, self).get_context_data(**kwargs)

        object_id = self.get_object().id
        setattr(self.request, 'productid', object_id)

        return context

    def get_object(self, queryset=None):
        try:
            if queryset is None:
                queryset = self.model.objects.filter(published=True)
            product = queryset.get(slug=self.kwargs['product_slug'])
            setattr(self.request, 'product', product)
            return product
        except Product.DoesNotExist:
            raise Http404


class ProductVariantDetailView(DetailView):
    model = ProductVariant
    context_object_name = 'variant'
    template_name = 'catalog/variant.html'

    def get_context_data(self, **kwargs):
        context = super(ProductVariantDetailView, self).get_context_data(**kwargs)

        object = self.get_object()
        object_id = object.id

        setattr(self.request, 'productid', object_id)

        return context

    def get_object(self, queryset=None):
        try:
            if queryset is None:
                queryset = self.model.objects.filter(product__published=True)
            return queryset.get(slug=self.kwargs['variant_slug'], product__slug=self.kwargs['product_slug'])
        except Product.DoesNotExist:
            raise Http404


def collections_list_view(request, template_name='catalog/collections.html'):
    collections = Collection.objects.filter(published=True)
    context = {
        'collections': collections,
        'page_url': request.path,
    }
    return render(request, template_name, context)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 18:41
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0031_vendor_description'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='productstatus',
            options={'ordering': ['-weight'], 'verbose_name': 'Product status', 'verbose_name_plural': 'Product statuses'},
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-26 17:18
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0012_remove_collection_exclude_properties_filters'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='exclude_properties_filters',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=70), blank=True, default=list, help_text='Comma separated list of properties labels which you want to exclude from filters.', size=None, verbose_name='Exclude properties filters'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-26 16:53
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0009_collectionpluginmodel'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='exclude_properties_filters',
            field=models.CharField(default='', help_text='Comma separated list of properties tags which you want to exclude from filters.', max_length=255, verbose_name='Exclude properties filters'),
        ),
        migrations.AddField(
            model_name='collection',
            name='show_price_filter',
            field=models.BooleanField(default=True, verbose_name='Show price filter'),
        ),
        migrations.AddField(
            model_name='collection',
            name='show_properties_filters',
            field=models.BooleanField(default=True, verbose_name='Show properties filters'),
        ),
        migrations.AddField(
            model_name='collection',
            name='show_vendor_filter',
            field=models.BooleanField(default=True, verbose_name='Show vendor filter'),
        ),
        migrations.AlterField(
            model_name='collection',
            name='url_filter_rules',
            field=models.TextField(blank=True, default='VENDOR\nPRICE', help_text='Use separate lines for different rules. See documentation for details.', verbose_name='URL filter rules'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-02 07:54
from __future__ import unicode_literals

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('pcart_catalog', '0068_collectionpluginmodel_ajax_load'),
    ]

    operations = [
        migrations.CreateModel(
            name='SimilarProductsPluginModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='pcart_catalog_similarproductspluginmodel', serialize=False, to='cms.CMSPlugin')),
                ('show_title', models.BooleanField(default=True, verbose_name='Show title')),
                ('title', models.CharField(blank=True, default='', help_text='Optional. If not set the collection title will be used.', max_length=255, verbose_name='Title')),
                ('tags_or_prefixes', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=30), blank=True, default=list, help_text='You can specify like full tag as `prefix:value` but also `prefix` only.', size=None, verbose_name='Tags or prefixes')),
                ('vendor_filtering', models.BooleanField(default=False, verbose_name='Vendor filtering')),
                ('price_difference_filtering', models.CharField(blank=True, choices=[('', 'Do not filter'), ('absolute', 'Absolute difference'), ('relative', 'Relative difference')], default='', max_length=30, verbose_name='Price difference filtering')),
                ('price_filtering_value', models.DecimalField(decimal_places=2, default=0.0, max_digits=10, verbose_name='Price filtering value')),
                ('sorting', models.CharField(blank=True, default='', max_length=100, verbose_name='Sorting')),
                ('limit', models.PositiveIntegerField(default=6, verbose_name='Limit')),
                ('template_name', models.CharField(blank=True, default='', max_length=100, verbose_name='Template name')),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Optional. Some templates may support additional options for conditional rendering.', verbose_name='Data')),
                ('ajax_load', models.BooleanField(default=False, verbose_name='AJAX load')),
                ('collection', mptt.fields.TreeForeignKey(on_delete=django.db.models.deletion.CASCADE, to='pcart_catalog.Collection', verbose_name='Collection')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-09 11:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0007_auto_20170309_1144'),
    ]

    operations = [
        migrations.AddField(
            model_name='productimage',
            name='downloaded',
            field=models.BooleanField(default=False, verbose_name='Downloaded'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-10 12:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0020_auto_20170808_1104'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='custom_url_filter_rules',
            field=models.BooleanField(default=False, verbose_name='Custom URL filter rules'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-04-11 07:25
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0013_collection_exclude_properties_filters'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionpluginmodel',
            name='collection',
            field=mptt.fields.TreeForeignKey(on_delete=django.db.models.deletion.CASCADE, to='pcart_catalog.Collection', verbose_name='Collection'),
        ),
        migrations.AlterField(
            model_name='collectionpluginmodel',
            name='filter_string',
            field=models.CharField(blank=True, default='', help_text='Use the list of filter chunks separated with / and ; characters.', max_length=255, verbose_name='Filter string'),
        ),
        migrations.AlterField(
            model_name='collectionpluginmodel',
            name='sorting',
            field=models.CharField(blank=True, default='', max_length=100, verbose_name='Sorting'),
        ),
        migrations.AlterField(
            model_name='collectionpluginmodel',
            name='template_name',
            field=models.CharField(blank=True, default='', max_length=100, verbose_name='Template name'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-29 13:33
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0040_migrate_translatable_fields_for_productstatus'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='productstatus',
            name='title',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-08 08:04
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0019_auto_20170505_1202'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='external_id',
            field=models.CharField(blank=True, db_index=True, default='', max_length=255, verbose_name='External ID'),
        ),
        migrations.AddField(
            model_name='product',
            name='quantity',
            field=models.PositiveIntegerField(blank=True, default=None, null=True, verbose_name='Quantity'),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='external_id',
            field=models.CharField(blank=True, db_index=True, default='', max_length=255, verbose_name='External ID'),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='quantity',
            field=models.PositiveIntegerField(blank=True, default=None, null=True, verbose_name='Quantity'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-20 18:47
from __future__ import unicode_literals

import adminsortable.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0004_auto_20170220_2032'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='productimage',
            options={'ordering': ['position'], 'verbose_name': 'Product image', 'verbose_name_plural': 'Product images'},
        ),
        migrations.AlterField(
            model_name='productimage',
            name='position',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False, verbose_name='Position'),
        ),
        migrations.AlterField(
            model_name='productimage',
            name='product',
            field=adminsortable.fields.SortableForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images', to='pcart_catalog.Product', verbose_name='Product'),
        ),
        migrations.AlterField(
            model_name='producttypeproperty',
            name='position',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False, verbose_name='Position'),
        ),
        migrations.AlterField(
            model_name='producttypeproperty',
            name='product_type',
            field=adminsortable.fields.SortableForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='pcart_catalog.ProductType', verbose_name='Product type'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 12:40
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0023_auto_20170927_1535'),
    ]

    operations = [
        migrations.AddField(
            model_name='producttypeproperty',
            name='hidden',
            field=models.BooleanField(default=False, help_text='Useful for hide some product specified information from a web page.', verbose_name='Hidden'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-27 17:03
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pcart_catalog', '0025_auto_20170927_1959'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='products_selection',
            field=models.CharField(choices=[('manual', 'Manual'), ('automatic', 'Automatic')], default='manual', help_text='Use Automatic mode if you want to select products based on conditions.', max_length=30, verbose_name='Products selection'),
        ),
    ]
